#include <stdio.h>
#include <stdint.h>
#include <i86.h>
#include <conio.h>
#include <dos.h>

#include "dpmi.h"
#include "fxmath.h"

#define arrayof(n) (sizeof(n) / sizeof(n[0]))

// set unchained aka mode-x
void setmodex() {
    // disable chain-4
    outp(0x3C4, 0x4);  outp(0x3C5, (inp(0x3C5) & ~0x08));
    // set crtc byte mode
    outp(0x3D4, 0x14); outp(0x3D5, (inp(0x3D5) & ~0x40));
    outp(0x3D4, 0x17); outp(0x3D5, (inp(0x3D5) |  0x40));
}

void setvgamode(int mode);
#pragma aux setvgamode = "int 0x10" parm [eax]

// IRQ0 interrupt handler
extern "C" uint32_t irq0_count = 0;
extern "C" uint32_t pit_getTickCount();
#if 0
void __declspec(naked) irq0_isr() {
    _asm {
        push    eax
        inc     dword ptr [irq0_count]
        mov     al, 0x20
        out     0x20, al
        pop     eax
        iret
    }
}
#else
void __far __interrupt irq0_isr() {
    irq0_count++;
    outp(0x20, 0x20);
}
#endif

// enable light pen position readback (the dreaded "PRNG" :D)
void lightpen_pos_enable() {
    outp(0x3d4, 0x11); outp(0x3d5, inp(0x3d5) & 0x7F);  // unlock registers
    outp(0x3d4, 0x03); outp(0x3d5, inp(0x3d5) & 0x7F);  // enable lightpen position readback
}

// read current screen address
int lightpen_pos_read(){
    int rtn = 0;
    _asm {
        mov     edx, 0x3d4
        xor     eax, eax
        xor     ebx, ebx
        mov     ecx, 0x1110

    _reread:
        // read first word
        mov     al, cl
        out     dx, al
        inc     edx
        in      al, dx
        dec     edx
        mov     ah, al
        mov     al, ch
        out     dx, al
        inc     edx
        in      al, dx
        dec     edx
        mov     ebx, eax         // EAX - current lightpen position

        // read high byte to check counter carry during halves reading (they are NOT latched!)
        mov     al, cl
        out     dx, al
        inc     edx
        in      al, dx
        cmp     bh, al
        jne     _reread

        // done :)
        mov     [rtn], eax
    }

    return rtn;
}

// assumes lighpen readback is enabled
bool lightpen_pos_check() {
    int timeout = 1024;             // max counter reads until bail out
    int ticks_to_pass = 32;         // number of counter changes needed to pass the check
    uint8_t oldval = 0;

    // check low half
    outp(0x3d4, 0x11); oldval = inp(0x3d5);
    while ((--timeout) && (ticks_to_pass != 0)) {
        uint8_t newval = inp(0x3d5);
        if (newval != oldval) {
            oldval = newval;
            ticks_to_pass--;
        }
    }

    return (timeout > 0);
}

// read data from I/O port and save to file
// returns ticks to read

uint32_t dump_port8(const char *filename, uint32_t port, uint8_t *buf, uint32_t size) {
    // prefetch data to warm up the cache
    _asm {
        mov     esi, [buf]
        mov     edi, esi
        mov     ecx, [size]
        shr     ecx, 2
        rep     movsd
    }

    // wait for vertical retrace
    while (inp(0x3da) & 8); while (!(inp(0x3da) & 8));

    // capture and time the frame
    uint32_t ticks = pit_getTickCount();
    _asm {
        mov     edi, [buf]
        mov     ecx, [size]
        mov     edx, [port]
        rep     insb
    }
    ticks = pit_getTickCount() - ticks;

    // write to the file
    FILE *f = fopen(filename, "wb");
    fwrite(buf, size, 1, f);
    fclose(f);

    return ticks;
}

struct mode_test_t {
    bool        show;
    const char *name;
    int         vgamode;
    int         start_address;

    const char  *port3da_fname;
    const char  *lppos_lo_fname;
    const char  *lppos_hi_fname;

    bool        lppos_active;
    uint32_t    port3da_ticks;
    uint32_t    lppos_lo_ticks;
    uint32_t    lppos_hi_ticks;

    uint8_t     cr10, cr11;
};

mode_test_t mode_test[] = {
    {true,  "320x200 16 color",  0x0D, 0x0000, "16c_3da.000",  "16c_lo.000",   "16c_hi.000"},
    {false, "320x200 16 color",  0x0D, 0x5555, "16c_3da.555",  "16c_lo.555",   "16c_hi.555"},
    {true,  "320x200 256 color", 0x13, 0x0000, "256c_3da.000", "256c_lo.000",  "256c_hi.000"},
    {false, "320x200 256 color", 0x13, 0x5555, "256c_3da.555", "256c_lo.555",  "256c_hi.555"},
    {true,  "320x200 256 color Mode-X", 0x93, 0x0000, "mx_3da.000", "mx_lo.000",  "mx_hi.000"},
    {false, "320x200 256 color Mode-X", 0x93, 0x5555, "mx_3da.555", "mx_lo.555",  "mx_hi.555"},
};

// convert ticks to kilobytes per second for given buffer size
uint32_t ticks_to_kbytes(uint32_t ticks, uint32_t bufsize) {
    return imuldiv(0x1234D, bufsize, ticks << 6);
}

bool lppen_mode_test(mode_test_t* mode_test, uint8_t* portbuf, uint32_t portbufsize) {
    setvgamode(mode_test->vgamode & 0x7F);
    if (mode_test->vgamode & 0x80) setmodex();
    
    // clear the screen
    outpw(0x3c4, 0x0F02);
    _asm {
        mov edi, 0xA0000
        mov ecx, 16384
        xor eax, eax
        rep stosd
    }

    // setup diagmux, feed bits 7/6
    inp(0x3da); outp(0x3c0, 0x32); outp(0x3c0, 0x3F);

    // set border color to 0xF0, so border color is feeded to the diagmux
    inp(0x3da); outp(0x3c0, 0x31); outp(0x3c0, 0xF0);

    // read source CR10/CR11 values
    outp(0x3d4, 0x10); mode_test->cr10 = inp(0x3d5);
    outp(0x3d4, 0x11); mode_test->cr11 = inp(0x3d5);

    // set display start
    outpw(0x3d4, ((mode_test->start_address >> 0) & 0xFF00) | 0xC);
    outpw(0x3d4, ((mode_test->start_address >> 8) & 0xFF00) | 0xD);

    // enable lightpen pos
    lightpen_pos_enable();

    // and check it
    mode_test->lppos_active = lightpen_pos_check();

    // wait for vertical retrace
    // first collect port 0x3DA statistics
    mode_test->port3da_ticks = dump_port8(mode_test->port3da_fname, 0x3DA, portbuf, portbufsize);

    // dump low byte
    outp(0x3d4, 0x11);
    mode_test->lppos_lo_ticks = dump_port8(mode_test->lppos_lo_fname, 0x3D5, portbuf, portbufsize);

    // dump high byte
    outp(0x3d4, 0x10);
    mode_test->lppos_hi_ticks = dump_port8(mode_test->lppos_hi_fname, 0x3D5, portbuf, portbufsize);

    return mode_test->lppos_active;
}

int main(int argc, char* argv[]) {
    uint32_t portbufsize = (1 << 17);
    if (argc >= 2) portbufsize = atoi(argv[1]) * 1024;

    // set timer mode 2 and reload 65536 (aka 18.2hz)
    outp(0x43, 0x34); outp(0x40, 0); outp(0x40, 0);

    // install custom IRQ0 ISR
    void __far *oldirq0 = dpmi_getpmvect(8);
    dpmi_setpmvect(8, irq0_isr);

    // allocate port buffer
    uint8_t* portbuf = new uint8_t[portbufsize];

    for (int i = 0; i < arrayof(mode_test); i++) {
        lppen_mode_test(mode_test + i, portbuf, portbufsize);
    }

    // back to the textmode
    setvgamode(0x3);

    // print the results
    mode_test_t *cm = mode_test;
    for (int i = 0; i < arrayof(mode_test); i++, cm++) if (cm->show) {
        printf("results for %s\n", cm->name);
        printf("VGA Light Pen Position Readback:    %s\n", (cm->lppos_active ? "PASS" : "FAIL"));
        printf("CRTC registers 0x10/11:             %02X / %02X\n", cm->cr10, cm->cr11);
        printf("Port 0x3DA readback time:           %-7d ticks - %5d kb/s\n", cm->port3da_ticks,  ticks_to_kbytes(cm->port3da_ticks,  portbufsize));
        printf("Light Pen Low  Pos readback time:   %-7d ticks - %5d kb/s\n", cm->lppos_lo_ticks, ticks_to_kbytes(cm->lppos_lo_ticks, portbufsize));
        printf("Light Pen High Pos readback time:   %-7d ticks - %5d kb/s\n", cm->lppos_hi_ticks, ticks_to_kbytes(cm->lppos_hi_ticks, portbufsize));
        printf("----------------------------------------\n");
    }
    printf("I/O Port Buffer size: %d bytes\n", portbufsize);

    // cleanup
    delete[] portbuf;
    dpmi_setpmvect(8, oldirq0);
    return 0;
}
